from trezorpass.store.sources import Source


class FileSource(Source):
    def __init__(self, filename: str | None) -> None:
        self.filename = filename

    async def load_store(self, store_name) -> bytes:
        filename = self.filename if self.filename else store_name
        with open(filename, "rb") as file:
            return file.read()



import json

from trezorpass.crypto import decrypt
from trezorpass.store.entry import EntryDecoder
from trezorpass.store.tag import TagDecoder
from trezorpass.store.containers import Store
from trezorpass.store.keychain import Keychain
from trezorpass.store.sources import Source


class StoreLoader:
    def __init__(self, keychain: Keychain):
        self.keychain = keychain

    async def load(self, source: Source) -> bytes:
        try:
            return await source.load_store(self.keychain.store_name)
        except Exception as e:
            raise StoreLoadError() from e


class StoreDecrypter:
    def __init__(self, keychain: Keychain):
        self.keychain = keychain

    def decrypt(self, encrypted_store: bytes) -> bytes:
        try:
            return decrypt(self.keychain.store_key, encrypted_store)
        except Exception as e:
            raise StoreDecryptError() from e


class StoreDecoder:
    def __init__(self, keychain: Keychain, entry_decoder: EntryDecoder, tag_decoder: TagDecoder):
        self.keychain = keychain
        self.entry_decoder = entry_decoder
        self.tag_decoder = tag_decoder

    def decode(self, encoded_store: bytes) -> Store:
        try:
            store_dict = json.loads(encoded_store)
            tags_dict = store_dict['tags']
            entries_dict = store_dict['entries']
            tags = [self.tag_decoder.decode(tags_dict[key]) for key in tags_dict]
            entries = [self.entry_decoder.decode(entries_dict[key]) for key in entries_dict]
            return Store(
                name=self.keychain.store_name,
                entries=entries,
                tags=tags
            )
        except Exception as e:
            raise StoreDecodeError() from e


class StoreManager:
    def __init__(self, source: Source, *, loader: StoreLoader, decrypter: StoreDecrypter, decoder: StoreDecoder):
        self.source = source
        self.loader = loader
        self.decrypter = decrypter
        self.decoder = decoder

    async def __aenter__(self):
        loaded_store = await self.loader.load(self.source)
        decrypted_store = self.decrypter.decrypt(loaded_store)
        self.store = self.decoder.decode(decrypted_store)
        return self.store

    async def __aexit__(self, *args):
        pass


class StoreLoadError(Exception):
    pass


class StoreDecryptError(Exception):
    pass


class StoreDecodeError(Exception):
    pass



import os

from InquirerPy.utils import color_print
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import FormattedText

PRIMARY_COLOR = "#00783d"
SECONDARY_COLOR = "#f7931a"

os.environ["INQUIRERPY_STYLE_QUESTIONMARK"] = PRIMARY_COLOR
os.environ["INQUIRERPY_STYLE_ANSWERMARK"] = PRIMARY_COLOR
os.environ["INQUIRERPY_STYLE_ANSWER"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_INPUT"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_POINTER"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_FUZZY_PROMPT"] = PRIMARY_COLOR


def prompt_print(message: str, **kwargs):
    print_formatted_text(
        FormattedText([(PRIMARY_COLOR, "# "), ("class:message", message)]),
        **kwargs
    )


def prompt_print_pairs(pairs: list[tuple[str, str]]):
    color_print([(SECONDARY_COLOR, "#")])
    for (name, value) in pairs:
        color_print([
            (PRIMARY_COLOR, "# "),
            ("class:name", f"{name}: "),
            (SECONDARY_COLOR, value if value else "<empty>")
        ])
    color_print([(SECONDARY_COLOR, "#")])


def welcome():
    banner = """####################################################
#                                                  #
#       #######                                    #
#     ##       ##                                  #
#    ##         ##                                 #
#    ##         ##                                 #
#  #################                               #
#  ##             ##                               #
#  ##             ##                               #
#  ##             ##  Trezor Password Manager CLI  #
#  ##             ##                               #
#  ##             ##                               #
#  ###           ###                               #
#    ###       ###                                 #
#      #### ####                                   #
#         ###                                      #
#                                                  #
####################################################"""
    color_print(prepare_graphics(banner), {"graphics": PRIMARY_COLOR, "text": SECONDARY_COLOR})


def goodbye():
    message = """##################### Goodbye ######################"""
    color_print(prepare_graphics(message), {"graphics": PRIMARY_COLOR, "text": SECONDARY_COLOR})


def pin_guide():
    guide = """#
#       #######                #######       
#     ##       ##            ##       ##   
#    ##         ##          ##         ##  
#    ##         ##          ##         ##  
#  #################      #################
#  ##  Input PIN  ##      ##  Input PIN  ## 
#  ##             ##      ##             ##
#  ##    7 8 9    ##      ##    e r t    ##
#  ##    4 5 6    ##  or  ##    d f g    ##
#  ##    1 2 3    ##      ##    c v b    ##
#  ###           ###      ###           ###
#    ###       ###          ###       ###  
#      #### ####              #### ####    
#         ###                    ###
#"""
    color_print(prepare_graphics(guide), {"graphics": PRIMARY_COLOR, "text": SECONDARY_COLOR})


def confirm_guide():
    guide = """#
#       #######     
#     ##       ##   
#    ##         ##  
#    ##         ##  
#  #################
#  ##             ##
#  ##   Confirm   ##
#  ##             ##
#  ##             ##
#  ##  No    Yes  ##
#  ###           ###
#    ###       ###  
#      #### ####    
#         ###
#"""
    color_print(prepare_graphics(guide), {"graphics": PRIMARY_COLOR, "text": SECONDARY_COLOR})


def animate_dots(max_dots: int = 5):
    """Indefinitely keeps printing dots into the terminal, wrapping at specified value

    Args:
        max_dots: number of dots to wrap after
    """
    dots = 0
    while True:
        yield
        if dots == max_dots:
            print("\b \b" * dots, end="", flush=True)
            dots = 0
        else:
            print(".", end="", flush=True)
            dots += 1


def prepare_graphics(graphics: str,
                     graphics_characters: str = "#",
                     graphics_class: str = "graphics",
                     text_class: str = "text"):
    """Prepares ASCII graphics to be printed by color_print

    Args:
        graphics: ASCII graphics
        graphics_characters: String of characters that are considered to be visual only
        graphics_class: Class to be used for visual characters
        text_class: Class to be used for non-visual characters
    """
    result = []
    buffer = ""
    is_graphics = True
    for c in graphics:
        if c.isspace():
            buffer += c
        elif is_graphics and c not in graphics_characters:
            result.append((f"class:{graphics_class}", buffer))
            is_graphics = False
            buffer = c
        elif not is_graphics and c in graphics_characters:
            result.append((f"class:{text_class}", buffer))
            is_graphics = True
            buffer = c
        else:
            buffer += c
    result.append((f"class:{graphics_class if is_graphics else text_class}", buffer))
    return result



from trezorpass.store.containers import *
from trezorpass.store.entry import *
from trezorpass.store.store import *
from trezorpass.store.tag import *


def get_store_manager(keychain: Keychain, source: Source) -> StoreManager:
    return StoreManager(
        source,
        loader=StoreLoader(keychain),
        decrypter=StoreDecrypter(keychain),
        decoder=StoreDecoder(keychain, EntryDecoder(), TagDecoder())
    )



import asyncio

from trezorlib.client import TrezorClient
from trezorlib.transport import TransportException

from trezorpass.client import get_default_client
from trezorpass.utils import animate_dots, prompt_print


async def get_client() -> TrezorClient:
    """Waits for the Trezor client to connect

    Returns:
        The connected Trezor client

    Raises:
        KeyboardInterrupt
        Exception: Client initialization failure
    """
    prompt_print("Looking for a Trezor device ", end="", flush=True)
    animation = animate_dots(5)
    while True:
        try:
            client = get_default_client()
            print()
            return client
        except TransportException:
            next(animation)  # Keep waiting
        except Exception:
            print()
            print("Unable to access a Trezor device")
            raise
        await asyncio.sleep(0.7)



import logging
import shutil
from pathlib import Path

import appdirs

APP_DIR = appdirs.user_data_dir('trezor-pass')


def init_data():
    try:
        Path(APP_DIR).mkdir(parents=True, exist_ok=True)
    except:
        logging.exception("Unable to initialize application data directory")


def clear_data():
    """Clears stored application data"""
    shutil.rmtree(APP_DIR, ignore_errors=True)


init_data()



import getpass
from typing import Optional, Union

from trezorlib import messages
from trezorlib.exceptions import Cancelled
from trezorlib.messages import PinMatrixRequestType
from trezorlib.client import MAX_PIN_LENGTH

from trezorpass.utils import pin_guide, confirm_guide, prompt_print


class ManagerUI:
    @staticmethod
    def button_request(br: messages.ButtonRequest) -> None:
        confirm_guide()

    @staticmethod
    def get_pin(code: Optional[PinMatrixRequestType]) -> str:
        pin_guide()
        try:
            prompt_print("", end="")
            pin = getpass.getpass("Please enter PIN: ")
        except KeyboardInterrupt:
            raise Cancelled

        if all(d in "cvbdfgert" for d in pin):
            pin = pin.translate(str.maketrans("cvbdfgert", "123456789"))

        if any(d not in "123456789" for d in pin):
            prompt_print(
                "The value may only consist of digits 1 to 9 or letters cvbdfgert."
            )
            raise Cancelled
        elif len(pin) > MAX_PIN_LENGTH:
            prompt_print(f"The value must be at most {MAX_PIN_LENGTH} digits in length.")
            raise Cancelled
        else:
            return pin

    @staticmethod
    def get_passphrase(available_on_device: bool) -> Union[str, object]:
        return ""



class Source:
    """Loads the password store
    
    Args:
        store_name: Name of the store to load, implementors may ignore this argument
    """
    async def load_store(self, store_name: str | None) -> bytes:
        raise NotImplementedError()


class SourceError(Exception):
    pass



import asyncio
import logging

from trezorlib.exceptions import PinException

from trezorpass.store import StoreLoadError, StoreDecryptError, StoreDecodeError, get_store_manager, EntryDecrypter, \
    Keychain
from trezorpass.store.sources import Source, DropboxSource, FileSource
from trezorpass.utils import prompt_print, welcome, goodbye
from trezorpass.appdata import clear_data
from trezorpass.helpers import get_client, select_entry, manage_entry


async def cli(store_source: Source):
    welcome()
    try:
        with await get_client() as client:
            keychain = Keychain(client)
            async with get_store_manager(keychain, store_source) as store:
                while True:
                    entry = await select_entry(store.entries)
                    await manage_entry(entry, EntryDecrypter(keychain))
    except KeyboardInterrupt:
        pass
    except asyncio.CancelledError:
        pass
    except PinException:
        prompt_print("Trezor pin was not valid")
    except (StoreLoadError, StoreDecryptError, StoreDecodeError):
        prompt_print("Failed to load the password store")
    except BaseException as e:
        logging.exception("CLI failed", exc_info=e)
    finally:
        goodbye()


def run():
    import argparse
    parser = argparse.ArgumentParser(description='Command line interface for interaction with Trezor password store.')
    parser.add_argument("--clear", action='store_true', help="clears saved application data")
    parser.add_argument("--store", type=str, help="specifies the store file to be used instead of remote store")
    parser.add_argument("--debug", action='store_true', help="print debug logs")
    args = parser.parse_args()

    if not args.debug:
        logging.disable()

    if args.clear:
        clear_data()
    elif args.store:
        asyncio.run(cli(FileSource(args.store)))
    else:
        asyncio.run(cli(DropboxSource()), debug=True if args.debug else False)


if __name__ == "__main__":
    run()



from hmac import HMAC
from hashlib import sha256
from urllib.parse import urlparse

from trezorlib.client import TrezorClient
from trezorlib.exceptions import TrezorFailure
from trezorlib.misc import encrypt_keyvalue, decrypt_keyvalue
from trezorlib.tools import parse_path

from trezorpass.crypto import PATH, FILENAME_MESS
from trezorpass.store.containers import Entry


class Keychain:
    def __init__(self, client: TrezorClient):
        self.client = client

        address_n = parse_path(PATH)
        key = "Activate TREZOR Password Manager?"
        value = bytes.fromhex(
            "2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee"
        )
        try:
            self.master_key = encrypt_keyvalue(client, address_n, key, value).hex()
        except Exception as e:
            raise TrezorFailure() from e

    @property
    def store_key(self) -> str:
        return self.master_key[len(self.master_key) // 2:]

    @property
    def store_name(self) -> str:
        file_key = self.master_key[:len(self.master_key) // 2]
        return HMAC(file_key.encode(), FILENAME_MESS.encode(), sha256).hexdigest() + '.pswd'

    def entry_key(self, entry: Entry) -> str:
        address_n = parse_path(PATH)
        url = urlparse(entry.url)
        if url.scheme in ('ftp', 'http', 'https'):
            domain = url.netloc
        else:
            domain = url
        key = f'Unlock {domain} for user {entry.username}?'
        value = bytes.fromhex(entry.nonce)
        try:
            return decrypt_keyvalue(self.client, address_n, key, value, ask_on_encrypt=False).hex()
        except Exception as e:
            raise TrezorFailure() from e



import logging
from datetime import datetime
import json
import os

import dropbox
from InquirerPy import inquirer

from trezorpass.store.sources.source import Source, SourceError
from trezorpass.appdata import APP_DIR

DROPBOX_APP_KEY = "s340kh3l0vla1nv"  # APP_KEY of the official TPM, potentially breaking if maintainers disable
# PKCE flow for Dropbox auth
DROPBOX_TOKEN_FILE = os.path.join(APP_DIR, 'dropbox')


class OAuth:
    def __init__(self, access_token: str = None, refresh_token: str = None, expiration: str = None):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expiration = expiration

    @staticmethod
    def load():
        oauth = OAuth()
        with open(DROPBOX_TOKEN_FILE, 'r') as file:
            oauth.__dict__ = json.load(file)
            return oauth

    def store(self):
        with open(DROPBOX_TOKEN_FILE, 'w+') as file:
            os.chmod(DROPBOX_TOKEN_FILE, 0o600)
            file.write(json.dumps(self.__dict__))


class DropboxSource(Source):
    def __init__(self):
        try:
            self.oauth = OAuth.load()
        except Exception as e:
            self.oauth = None

    @staticmethod
    async def authenticate():
        auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(DROPBOX_APP_KEY, use_pkce=True, token_access_type='offline')
        authorize_url = auth_flow.start()
        print("1. Go to: " + authorize_url)
        print("2. Click \"Allow\" (you might have to log in first).")
        print("3. Copy the authorization code.")
        oauth_result = None
        while not oauth_result:
            auth_code = await inquirer.text("Enter the authorization code here:").execute_async()
            try:
                oauth_result = auth_flow.finish(auth_code)
            except:
                retry = await inquirer.confirm("Invalid authorization code, retry?").execute_async()
                if not retry:
                    raise SourceError()
        oauth = OAuth(oauth_result.access_token, oauth_result.refresh_token, oauth_result.expires_at.isoformat())
        try:
            oauth.store()
        except Exception as ex:
            logging.exception("Unable to store the oauth tokens")
        return oauth

    async def load_store(self, store_name) -> bytes:
        if not self.oauth:
            self.oauth = await self.authenticate()
        with dropbox.Dropbox(
                oauth2_access_token=self.oauth.access_token,
                oauth2_refresh_token=self.oauth.refresh_token,
                oauth2_access_token_expiration=datetime.fromisoformat(self.oauth.expiration),
                app_key=DROPBOX_APP_KEY
        ) as dbx:
            (_, response) = dbx.files_download("/" + store_name)
            return response.content



from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

PATH = "m/10016'/0"
FILENAME_MESS = '5f91add3fa1c3c76e90c90a3bd0999e2bd7833d06a483fe884ee60397aca277a'
CIPHER_IVSIZE = 96 // 8
AUTH_SIZE = 128 // 8


def decrypt(key: str, data: bytes) -> bytes:
    """Decrypts data using AES-GCM.
    Used for decrypting the store and entry secrets.

    Args:
        key: Key for the data acquired from Trezor device
        data: Binary data containing the (iv + ciphertext + authtag)
    """
    iv = data[:CIPHER_IVSIZE]
    auth_tag = data[CIPHER_IVSIZE: CIPHER_IVSIZE + AUTH_SIZE]
    ciphertext = data[CIPHER_IVSIZE + AUTH_SIZE:]
    cipher = Cipher(algorithms.AES(bytes.fromhex(key)), modes.GCM(iv, auth_tag))
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()



import json

from trezorpass.crypto import decrypt
from trezorpass.store.containers import EncryptedEntry, DecryptedEntry
from trezorpass.store.keychain import Keychain


class EntryDecoder:
    def decode(self, entry_dict) -> EncryptedEntry:
        return EncryptedEntry(
            url=entry_dict["title"],  # Intended
            title=entry_dict["note"],  # Intended
            username=entry_dict["username"],
            nonce=entry_dict["nonce"],
            encrypted_password=entry_dict["password"]["data"],
            encrypted_safe_note=entry_dict["safe_note"]["data"],
            tags=[]
        )


class EntryDecrypter:
    def __init__(self, keychain: Keychain):
        self.keychain = keychain

    def decrypt(self, entry: EncryptedEntry) -> DecryptedEntry:
        key = self.keychain.entry_key(entry)
        password = json.loads(decrypt(key, bytes(entry.encrypted_password)).decode("utf8"))
        safe_note = json.loads(decrypt(key, bytes(entry.encrypted_safe_note)).decode("utf8"))
        return DecryptedEntry(
            url=entry.url,
            title=entry.title,
            username=entry.username,
            nonce=entry.nonce,
            tags=entry.tags,
            password=password,
            safe_note=safe_note
        )



from typing import List


class Tag:
    def __init__(self, *, title: str):
        self.title = title


class Entry:
    def __init__(self, *, url: str, title: str, username: str, nonce: str, tags: List[Tag]):
        self.url = url
        self.title = title
        self.username = username
        self.nonce = nonce
        self.tags: List[Tag] = tags


class EncryptedEntry(Entry):
    def __init__(self, *, encrypted_password: str, encrypted_safe_note: str, **kwargs):
        super().__init__(**kwargs)
        self.encrypted_password = encrypted_password
        self.encrypted_safe_note = encrypted_safe_note


class DecryptedEntry(Entry):
    def __init__(self, *, password: str, safe_note: str, **kwargs):
        super().__init__(**kwargs)
        self.password = password
        self.safe_note = safe_note


class Store:
    def __init__(self, *, name: str, entries: List[Entry], tags: List[Tag]):
        self.name = name
        self.entries = entries
        self.tags = tags



from typing import List
import webbrowser

from trezorlib.exceptions import Cancelled
from InquirerPy import inquirer
from pyperclip import copy

from trezorpass.store import Entry, EncryptedEntry, EntryDecrypter
from trezorpass.utils import prompt_print, prompt_print_pairs


async def select_entry(entries: List[Entry]) -> Entry:
    """Facilitates user interaction to select an entry from the given list of entries

    Returns:
        A single entry from the specified list of entries

    Raises:
        KeyboardInterrupt
    """
    choices = [{"value": entry, "name": entry.title} for entry in entries]
    selection = await inquirer.fuzzy(
        message="Select an entry:",
        choices=choices,
        long_instruction="Press Ctrl+C to exit"
    ).execute_async()
    return selection


async def manage_entry(entry: EncryptedEntry, decrypter: EntryDecrypter) -> None:
    """Facilitates interaction with the given entry"""
    clipboard_dirty = False
    try:
        while True:
            choices = [
                f'Open {entry.url}',
                'Copy username to clipboard',
                'Copy password to clipboard',
                'Show entry',
                'Show entry including secrets'
            ]
            action = await inquirer.select(
                "Select an action:",
                choices=choices,
                long_instruction="Press Ctrl+C to leave the entry"
                ).execute_async()
            try:
                if action == choices[0]:
                    webbrowser.open(entry.url, 2)
                elif action == choices[1]:
                    copy(entry.username)
                    prompt_print("Username has been copied to the clipboard")
                elif action == choices[2]:
                    decrypted_entry = decrypter.decrypt(entry)
                    copy(decrypted_entry.password)
                    clipboard_dirty = True
                    prompt_print("Password has been copied to the clipboard")
                elif action == choices[3]:
                    prompt_print_pairs([
                        ("URL", entry.url),
                        ("Title", entry.title),
                        ("Username", entry.username)
                    ])
                elif action == choices[4]:
                    decrypted_entry = decrypter.decrypt(entry)
                    prompt_print_pairs([
                        ("URL", decrypted_entry.url),
                        ("Title", decrypted_entry.title),
                        ("Username", decrypted_entry.username),
                        ("Password", decrypted_entry.password),
                        ("Safe Note", decrypted_entry.safe_note)
                    ])
            except Cancelled:
                prompt_print("Action has been cancelled")
    except KeyboardInterrupt:
        pass
    finally:
        if clipboard_dirty:
            prompt_print("Cleaning up the clipboard")
            copy("")



import threading
import asyncio
import logging

from trezorlib.client import TrezorClient
from trezorlib.transport import Transport, get_transport

from trezorpass.client.ui import ManagerUI


def get_default_client():
    """Creates default thread-safe client instance

    Returns:
        Instance of the thread-safe Trezor client

    Raises:
        TransportException: when client cannot be found
        Exception: when client initialization fails
    """
    transport = get_transport()
    ui = ManagerUI()
    return ExtendedTrezorClient(transport, ui)


class ExtendedTrezorClient(TrezorClient):
    def __init__(self, transport: Transport, ui: ManagerUI, **kwargs):
        self._lock = threading.Lock()
        self._healthcheck: asyncio.Task | None = None
        super().__init__(transport, ui, _init_device=False, **kwargs)

    def call_raw(self, msg):
        with self._lock:
            return super().call_raw(msg)

    def start_healthcheck(self, interval: float = 1):
        observer = asyncio.current_task()

        async def healthcheck_task():
            try:
                while self._healthcheck:
                    await asyncio.get_event_loop().run_in_executor(None, self.ping, "healthcheck")
                    await asyncio.sleep(interval)
            except Exception as e:
                observer.cancel()
                logging.exception("Client healthcheck has failed", exc_info=e)

        self._healthcheck = asyncio.get_event_loop().create_task(healthcheck_task())

    def clear_healthcheck(self):
        self._healthcheck = None

    def __enter__(self):
        self.init_device()
        self.start_healthcheck()
        return self

    def __exit__(self, *args):
        self.clear_healthcheck()
        self.end_session()
